SystemOrganization addCategory: #'AlgebraicDataType-Tests'!
SystemOrganization addCategory: #AlgebraicDataType!
SystemOrganization addCategory: #'AlgebraicDataType-Trees'!

Error subclass: #UnificationFailure
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType'!

!UnificationFailure commentStamp: 'fbs 5/24/2011 17:35' prior: 0!
I represent the inability to unify two patterns.!

TestCase subclass: #EmptyTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType-Tests'!

!EmptyTest methodsFor: 'testing' stamp: 'fbs 5/25/2011 07:38'!
testEmptyTreeHasNoChildren
	self deny: Empty new hasChildren.
	self deny: Tree empty hasChildren.
	
	self assert: Tree empty children isEmpty.! !

!EmptyTest methodsFor: 'testing' stamp: 'fbs 5/25/2011 22:33'!
testEmptyTreeIsNotVariable
	self deny: Empty new isMetaVariable.
	self deny: Tree empty isMetaVariable.! !

!EmptyTest methodsFor: 'testing' stamp: 'fbs 5/22/2011 21:09'!
testEqualsTestsStructuralEquality
	| t1 t2 t3 |
	t1 := Empty new.
	t2 := Empty new.
	t3 := t1 unapply.
	self deny: t1 == t2. "Not the same object."
	self assert: t1 = t2. "But are equal objects."
	self assert: t1 = t3. "Equality means structural equality."! !

!EmptyTest methodsFor: 'testing' stamp: 'fbs 5/22/2011 21:27'!
testUnapplyShouldDestructure
	self assert: Tree empty unapply = {#Empty}.! !

TestCase subclass: #LeafTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType-Tests'!

!LeafTest methodsFor: 'testing' stamp: 'fbs 5/22/2011 21:15'!
testEqualsTestsStructuralEquality
	| t1 t2 t3 |
	t1 := Leaf value: 1.
	t2 := Leaf value: 1.
	t3 := t1 unapply.
	self deny: t1 == t2. "Not the same object."
	self assert: t1 = t2. "But are equal objects."
	self assert: t1 = t3. "Equality means structural equality."! !

!LeafTest methodsFor: 'testing' stamp: 'fbs 5/24/2011 22:07'!
testLeafHasChildren
	self assert: (Leaf value: 1) hasChildren.
	
	self assert: (Leaf value: 1) children = #(1).! !

!LeafTest methodsFor: 'testing' stamp: 'fbs 5/25/2011 22:32'!
testLeafIsNotVariable
	self deny: (Leaf value: 1) isMetaVariable.! !

!LeafTest methodsFor: 'testing' stamp: 'fbs 5/22/2011 21:15'!
testUnapplyShouldDestructure
	| destructure tree |
	tree := Leaf value: 1.
	destructure := tree unapply.
	self assert: destructure first = #Leaf.
	self assert: destructure second = tree value.! !

TestCase subclass: #MetaVariableTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType-Tests'!

!MetaVariableTest methodsFor: 'testing' stamp: 'fbs 5/24/2011 11:03'!
testUnapplyShouldDestructureCorrectly
	self assert: (MetaVariable named: #a) unapply = {#Variable. #a.}! !

!MetaVariableTest methodsFor: 'testing' stamp: 'fbs 5/24/2011 22:08'!
testVariableHasNoChildren
	self deny: (MetaVariable named: #foo) hasChildren.

	self assert: (MetaVariable named: #foo) children isEmpty.! !

!MetaVariableTest methodsFor: 'testing' stamp: 'fbs 5/25/2011 22:32'!
testVariableIsVariable
	self assert: (MetaVariable named: #foo) isMetaVariable.! !

TestCase subclass: #NodeTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType-Tests'!

!NodeTest methodsFor: 'testing' stamp: 'fbs 5/22/2011 20:58'!
testEqualsTestsStructuralEquality
	| t1 t2 t3 |
	t1 := Node left: Tree empty.
	t2 := Node left: Tree empty.
	t3 := t1 unapply.
	self deny: t1 == t2. "Not the same object."
	self assert: t1 = t2. "But are equal objects."
	self assert: t1 = t3. "Equality means structural equality."! !

!NodeTest methodsFor: 'testing' stamp: 'fbs 5/24/2011 22:08'!
testNodeHasChildren
	self assert: (Node left: Tree empty right: Tree empty) hasChildren.
	
	self assert: (Node left: (Leaf value: 1) right: (Leaf value: 2)) children = {Leaf value: 1. Leaf value: 2}.! !

!NodeTest methodsFor: 'testing' stamp: 'fbs 5/25/2011 22:32'!
testNodeIsNotVariable
	self deny: (Node left: Tree empty right: Tree empty) isMetaVariable.! !

!NodeTest methodsFor: 'testing' stamp: 'fbs 5/22/2011 20:59'!
testUnapplyShouldDestructureLazily
	| destructure tree |
	tree := Node left: (Leaf value: 1) right: (Leaf value: 2).
	destructure := tree unapply.
	self assert: destructure first = #Node.
	self assert: destructure second = tree left.
	self assert: destructure third = tree right.! !

TestCase subclass: #UnificationTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType-Tests'!

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/29/2011 15:09'!
testCyclicStructuresCantUnify
	| y |
	y := MetaVariable named: #y.
	self should: [y unifyWith: (Leaf value: y)] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/27/2011 20:28'!
testDifferentArityTermsShouldNotUnify
	"f(x) ~= g(x,y)"
	|f g |
	f := VariadicTerm with: #(#x).
	g := VariadicTerm with: #(#x #y).
	self should: [f unifyWith: g] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/29/2011 15:09'!
testDifferentConstantsShouldNotUnify
	self should: [(Leaf value: 1) unifyWith: (Leaf value: 2)] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/29/2011 15:06'!
testImpossibleUnificationShouldNotUnify
	self should: [(Leaf value: 2) unifyWith: (Leaf value: 1)] raise: UnificationFailure.
	self should: [1 unifyWith: 2] raise: UnificationFailure.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/29/2011 15:08'!
testSelfUnificationShouldUnify
	| x |
	x := MetaVariable named: #x.
	"Tautologies"
	self assert: (1 unifyWith: 1) isEmpty.
	self assert: (x unifyWith: x) isEmpty.! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/26/2011 07:43'!
testTrivialBidirectionalUnificationShouldUnify
	| matches x y |
	x := MetaVariable named: #x.
	y := MetaVariable named: #y.
	matches := x unifyWith: y.
	self assert: matches = (Dictionary newFrom: {x -> y}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/29/2011 15:08'!
testTrivialUnificationsShouldUnify
	| matches x leaf |
	leaf := MetaVariable named: #leaf.
	matches := leaf unifyWith: (Leaf value: 1).
	self assert: matches = (Dictionary newFrom: {leaf -> (Leaf value: 1)}).
	
	x := MetaVariable named: #x.
	self assert: (x unifyWith: 2) = (Dictionary newFrom: {x -> 2}).
	self assert: (2 unifyWith: x) = (Dictionary newFrom: {x -> 2}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/29/2011 15:11'!
testTwoVariableBidirectionalUnificationShouldUnifyBoth
	"node(left, leaf(2)) =? node(leaf(1), right)"
	| matches left right |
	left := MetaVariable named: #left.
	right := MetaVariable named: #right.
	matches := (Node left: left right: (Leaf value: 2))
		unifyWith: (Node left: (Leaf value: 1) right: right).
	self assert: matches = (Dictionary newFrom: {
		left -> (Leaf value: 1).
		right -> (Leaf value: 2)}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/26/2011 07:51'!
testTwoVariableUnificationShouldUnifyBoth
	| matches left right |
	left := MetaVariable named: #left.
	right := MetaVariable named: #right.
	matches := (Node left: left right: right)
		unifyWith: (Node left: (Leaf value: 1) right: (Leaf value: 2)).
	self assert: matches = (Dictionary newFrom: {
		left -> (Leaf value: 1).
		right -> (Leaf value: 2)}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/26/2011 07:52'!
testUnificationWorksOnDeepStructures
	| matches x y |
	x := MetaVariable named: #x.
	y := MetaVariable named: #y.
	matches := (Node left: (Node left: (Node left: x))) unifyWith: (Node left: y).
	self assert: matches = (Dictionary newFrom: {y -> (Node left: (Node left: x))}).! !

!UnificationTest methodsFor: 'testing' stamp: 'fbs 5/26/2011 08:24'!
testVariableUnifyingAgainstTwoVariables
	| matches x y z |
	x := MetaVariable named: #x.
	y := MetaVariable named: #y.
	z := MetaVariable named: #z.
	matches := (Node left: x right: x) unifyWith: (Node left: y right: z).
	self assert: matches = (Dictionary newFrom: {y -> z. x -> z}).! !

Object subclass: #AlgebraicDataType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType'!

!AlgebraicDataType commentStamp: 'fbs 5/25/2011 07:30' prior: 0!
My subclasses represent algebraic types. My subclasses know how to destructure themselves through implementing #unapply.

Further, for unification between two graphs of instances of my subclasses, you may think of the instances as functions. For instance, Node left: a right: b is conceptually the same as some function Node(left, right) taking Trees as parameters. Thus, you may access a structure's parts through a parameter-like API in #paramAt:, #numArgs, etc.!

!AlgebraicDataType methodsFor: 'comparing' stamp: 'fbs 5/22/2011 21:14'!
= anObject
	^ (anObject respondsTo: #unapply)
		ifTrue: [self unapply = anObject unapply]
		ifFalse: [anObject isCollection
			ifTrue: [self unapply = anObject]
			ifFalse: [false]]! !

!AlgebraicDataType methodsFor: 'testing' stamp: 'fbs 5/27/2011 22:06'!
arity
	^ self children size.! !

!AlgebraicDataType methodsFor: 'accessing' stamp: 'fbs 5/24/2011 22:06'!
children
	^ self unapply allButFirst.! !

!AlgebraicDataType methodsFor: 'testing' stamp: 'fbs 5/24/2011 22:04'!
hasChildren
	"Return true if self contains substructures/is of the form f(x1, x2, ..., xn)."
	"This covers most bases, but see MetaVariable for an exception."
	^ self unapply size > 1.! !

!AlgebraicDataType methodsFor: 'testing' stamp: 'fbs 5/27/2011 11:25'!
hasVariable
	"Return true if any element in my structure is a MetaVariable."
	^ self unapply anySatisfy: [:each | each isAlgebraic and: [each hasVariable]].! !

!AlgebraicDataType methodsFor: 'comparing' stamp: 'fbs 5/25/2011 22:54'!
hash
	"17's a prime."
	^ self unapply inject: 17 into: [:acc :each | 17 * each hash].! !

!AlgebraicDataType methodsFor: 'testing' stamp: 'fbs 5/25/2011 22:34'!
isAlgebraic
	^ true.! !

!AlgebraicDataType methodsFor: 'testing' stamp: 'fbs 5/25/2011 22:31'!
isMetaVariable
	^ self unapply first = #Variable.! !

!AlgebraicDataType methodsFor: 'accessing' stamp: 'fbs 5/25/2011 23:06'!
type
	^ self unapply first.! !

!AlgebraicDataType methodsFor: 'destructuring' stamp: 'fbs 5/22/2011 20:29'!
unapply
	self subclassResponsibility.! !

!AlgebraicDataType methodsFor: 'comparing' stamp: 'fbs 5/25/2011 07:12'!
unifyWith: anAlgebraicDataType
	^ Unifier unify: self with: anAlgebraicDataType.! !

AlgebraicDataType subclass: #MetaVariable
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType'!

!MetaVariable commentStamp: 'fbs 5/24/2011 11:04' prior: 0!
A MetaVariable is a reified variable, that is, an object representing a variable. It is a placeholder in a structure that says "we don't know what the value is of this part of the structure, but we can give it a name."!

!MetaVariable class methodsFor: 'instance creation' stamp: 'fbs 5/24/2011 07:31'!
named: aSymbol
	^ self new name: aSymbol.! !

!MetaVariable methodsFor: 'accessing' stamp: 'fbs 5/24/2011 22:06'!
children
	^ #().! !

!MetaVariable methodsFor: 'testing' stamp: 'fbs 5/24/2011 22:05'!
hasChildren
	"Even though unapplied a variable looks like it has a parameter, that's just its name."
	^ false.! !

!MetaVariable methodsFor: 'testing' stamp: 'fbs 5/27/2011 11:25'!
hasVariable
	^ true.! !

!MetaVariable methodsFor: 'accessing' stamp: 'fbs 5/24/2011 07:30'!
name
	^ name.! !

!MetaVariable methodsFor: 'private' stamp: 'fbs 5/24/2011 07:32'!
name: aSymbol
	name := aSymbol.! !

!MetaVariable methodsFor: 'destructuring' stamp: 'fbs 5/24/2011 07:34'!
unapply
	^ {#Variable. name.}! !

AlgebraicDataType subclass: #Tree
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType-Trees'!

Tree class
	instanceVariableNames: 'EmptyTree'!

Tree class
	instanceVariableNames: 'EmptyTree'!

Tree subclass: #Empty
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType-Trees'!

!Empty methodsFor: 'destructuring' stamp: 'fbs 5/22/2011 21:05'!
unapply
	^ #(#Empty).! !

Tree subclass: #Leaf
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType-Trees'!

!Leaf class methodsFor: 'instance creation' stamp: 'fbs 5/22/2011 20:19'!
value: anObject
	^ self new value: anObject.! !

!Leaf methodsFor: 'destructuring' stamp: 'fbs 5/22/2011 21:16'!
unapply
	^ {#Leaf. value.}! !

!Leaf methodsFor: 'accessing' stamp: 'fbs 5/22/2011 20:20'!
value
	^ value.! !

!Leaf methodsFor: 'private' stamp: 'fbs 5/22/2011 20:19'!
value: anObject
	value := anObject.! !

Tree subclass: #Node
	instanceVariableNames: 'left right'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType-Trees'!

!Node class methodsFor: 'instance creation' stamp: 'fbs 5/22/2011 20:21'!
left: aLeftObject 
	^ self left: aLeftObject right: Tree empty.! !

!Node class methodsFor: 'instance creation' stamp: 'fbs 5/22/2011 20:21'!
left: aLeftObject right: aRightObject
	^ self new left: aLeftObject right: aRightObject.! !

!Node class methodsFor: 'instance creation' stamp: 'fbs 5/22/2011 20:21'!
right: aRightObject
	^ self left: Tree empty right: aRightObject.! !

!Node methodsFor: 'accessing' stamp: 'fbs 5/22/2011 20:22'!
left
	^ left.! !

!Node methodsFor: 'private' stamp: 'fbs 5/22/2011 20:22'!
left: aLeftObject right: aRightObject
	left := aLeftObject.
	right := aRightObject.! !

!Node methodsFor: 'accessing' stamp: 'fbs 5/22/2011 20:22'!
right
	^ right.! !

!Node methodsFor: 'destructuring' stamp: 'fbs 5/22/2011 20:51'!
unapply
	^ {#Node. left. right.}! !

!Tree class methodsFor: 'accessing' stamp: 'fbs 5/22/2011 20:27'!
empty
	^ EmptyTree.! !

!Tree class methodsFor: 'class initialization' stamp: 'fbs 5/22/2011 20:28'!
initialize
	EmptyTree := Empty new.! !

AlgebraicDataType subclass: #VariadicTerm
	instanceVariableNames: 'subterms'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType-Tests'!

!VariadicTerm class methodsFor: 'instance variables' stamp: 'fbs 5/27/2011 20:26'!
with: aCollection
	^ self new with: aCollection.! !

!VariadicTerm methodsFor: 'destructuring' stamp: 'fbs 5/27/2011 20:27'!
unapply
	^ #Term asOrderedCollection, subterms.! !

!VariadicTerm methodsFor: 'private' stamp: 'fbs 5/27/2011 20:26'!
with: aCollection
	subterms := aCollection! !

!Object methodsFor: '*AlgebraicDataType' stamp: 'fbs 5/25/2011 22:34'!
isAlgebraic
	^ false.! !

!Object methodsFor: '*AlgebraicDataType' stamp: 'fbs 5/25/2011 22:33'!
isMetaVariable
	^ false.! !

!Object methodsFor: '*AlgebraicDataType' stamp: 'fbs 5/26/2011 20:42'!
unifyWith: anObject
	^ Unifier unify: self with: anObject.! !

Object subclass: #TermRelation
	instanceVariableNames: 'map'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType'!

!TermRelation commentStamp: 'fbs 5/28/2011 15:43' prior: 0!
I represent the term relation we build up as part of unifying two structures.!

!TermRelation methodsFor: 'accessing' stamp: 'fbs 5/28/2011 18:50'!
classFor: anObject
	^ map
		at: anObject
		ifAbsentPut: [UnificationEquivalenceClass decorate: anObject].! !

!TermRelation methodsFor: 'initialize-release' stamp: 'fbs 5/28/2011 15:52'!
initialize
	super initialize.
	map := Dictionary new.! !

Object subclass: #UnificationEquivalenceClass
	instanceVariableNames: 'classPointer classSize schemaTerm visited acyclic vars node'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType'!

!UnificationEquivalenceClass commentStamp: 'fbs 5/28/2011 15:45' prior: 0!
I decorate the nodes undergoing unification. Unifier associates a node with a UnificationDecoration to allow the unification state to remain separate from the immutable nodes.

* classPointer points to my equivalence class; if classPointer = self then I am the representative for my class.
* classSize is an optimisation: when we merge two equivalence classes, we merge the smaller class with the larger class.
* node points to the thing that I decorate.!

!UnificationEquivalenceClass class methodsFor: 'instance creation' stamp: 'fbs 5/25/2011 21:16'!
decorate: anObject
	^ self new decorate: anObject.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/28/2011 15:46'!
classPointer: anEquivClass
	classPointer := anEquivClass.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/25/2011 17:58'!
classSize
	^ classSize.! !

!UnificationEquivalenceClass methodsFor: 'initialize-release' stamp: 'fbs 5/25/2011 22:31'!
decorate: aNode
	self node: aNode.
	vars := aNode isMetaVariable
		ifTrue: [aNode asOrderedCollection]
		ifFalse: [OrderedCollection new].
	
	acyclic := false.
	classPointer := self. "All nodes start as the sole element in an equivalence class; this makes each node a representative of its class."
	schemaTerm := self.
	classSize := 1.
	visited := false.! !

!UnificationEquivalenceClass methodsFor: 'unification' stamp: 'fbs 5/29/2011 13:44'!
findRepresentative
	"Walk up the tree representing this equivalence class to find the representative. As a side effect, compress the path to the representative."
	| result |
	self isRepresentative ifTrue: [^ self].
	
	result := classPointer findRepresentative.
	classPointer := result.
	^ result.! !

!UnificationEquivalenceClass methodsFor: 'unification' stamp: 'fbs 5/29/2011 13:39'!
findSolutionWithUnifier: aDictionary in: termRelation
	^ (self isSchemaTerm)
		ifTrue: [self privFindSolutionWithUnifier: aDictionary in: termRelation]
		ifFalse: [self findRepresentative schemaTerm findSolutionWithUnifier: aDictionary in: termRelation]! !

!UnificationEquivalenceClass methodsFor: 'testing' stamp: 'fbs 5/29/2011 13:44'!
isRepresentative
	^ classPointer = self! !

!UnificationEquivalenceClass methodsFor: 'testing' stamp: 'fbs 5/29/2011 13:39'!
isSchemaTerm
	^ self = self findRepresentative schemaTerm.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/25/2011 20:07'!
node
	^ node.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/25/2011 20:07'!
node: anObject
	node := anObject.! !

!UnificationEquivalenceClass methodsFor: 'unification' stamp: 'fbs 5/29/2011 13:36'!
privFindSolutionWithUnifier: aDictionary in: termRelation
	"Assume that we are the schema term for our equivalence class."
	| newUnifier |
	acyclic ifTrue: [^aDictionary "s is not part of a cycle."].
	
	visited ifTrue: [UnificationFailure signal. "Cycle detected."].

	newUnifier := aDictionary copy.	
	(node isAlgebraic and: [node hasChildren]) ifTrue:
		[visited := true.
		node children do:
			[:each | newUnifier addAll: ((termRelation classFor: each) findSolutionWithUnifier: newUnifier in: termRelation)].
		visited := false].
	
	acyclic := true.
	self findRepresentative vars do:
		[:x | (x ~= node) ifTrue: [newUnifier at: x put: node]].
	
	^ newUnifier.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/25/2011 17:58'!
schemaTerm
	^ schemaTerm.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/28/2011 15:47'!
schemaTerm: anEquivClass
	schemaTerm := anEquivClass.! !

!UnificationEquivalenceClass methodsFor: 'unification' stamp: 'fbs 5/29/2011 13:46'!
unificationClosureWith: rightEquivClass in: termRelation
	| left right sTerm t tTerm |
	t := rightEquivClass findRepresentative.
	self isRepresentative ifFalse: [^ self findRepresentative unificationClosureWith: t in: termRelation].

	(self = t) ifTrue: [^ self."do nothing"].
		
	sTerm := schemaTerm.
	tTerm := t schemaTerm.
	left := sTerm node.
	right := tTerm node.
	
	(left isMetaVariable or: [right isMetaVariable])
		ifTrue: [^ self union: t].

	"s, t are both terms."
	(left isAlgebraic and: [right isAlgebraic])
		ifTrue:
			["s = f(s1, ..., sn), t = g(t1, ..., tn) where m, n > 0"
			(left type = right type) ifFalse: [UnificationFailure signal. "f ~= g"].
			(left arity = right arity) ifFalse: [UnificationFailure signal].
			sTerm union: tTerm.
			left children with: right children do:
				[:si :ti | (termRelation classFor: si) unificationClosureWith: (termRelation classFor: ti) in: termRelation]]
		ifFalse:
			["At least one of s, t are not AlgebraicDataTypes, but some other kind of value."
			(left = right)
				ifTrue: [sTerm union: tTerm]
				ifFalse: ["s ~= t" UnificationFailure signal]].! !

!UnificationEquivalenceClass methodsFor: 'unification' stamp: 'fbs 5/29/2011 13:45'!
union: anEquivClass
	"Given myself, a representative, and aDecorator, another representative, merge the two equivalence classes by folding the smaller into the larger."
	(self classSize >= anEquivClass classSize)
		ifFalse: [^ anEquivClass union: self].
	
	classSize := classSize + anEquivClass classSize.
	vars addAll: anEquivClass vars.
	schemaTerm node isMetaVariable ifTrue: [schemaTerm := anEquivClass schemaTerm].
	anEquivClass classPointer: self.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/25/2011 19:47'!
vars
	^ vars! !

Object subclass: #Unifier
	instanceVariableNames: 'termRelation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AlgebraicDataType'!

!Unifier commentStamp: 'fbs 5/25/2011 17:36' prior: 0!
I attempt to unify two structures, that is, find mappings between the MetaVariables (if any) within the two structures such that the structures represent the same structure.

I return a Dictionary containing the most general unifier.

classes holds the equivalence classes we build up as part of our term relation. It's a Dictionary mapping nodes to parent nodes.
sigma holds our unifications in triangular form.
vars is a Dictionary mapping the representative of each equivalence class to the variables in that class.!

!Unifier class methodsFor: 'instance creation' stamp: 'fbs 5/25/2011 07:12'!
unify: left with: right
	^ self new unify: left with: right.! !

!Unifier methodsFor: 'private' stamp: 'fbs 5/28/2011 19:51'!
findSolution: anEquivClass withUnifier: aDictionary
	^ anEquivClass findSolutionWithUnifier: aDictionary in: termRelation.! !

!Unifier methodsFor: 'evaluating' stamp: 'fbs 5/28/2011 19:53'!
unify: s with: t
	termRelation := TermRelation new.
	(termRelation classFor: s) unificationClosureWith: (termRelation classFor: t) in: termRelation.
	^ (termRelation classFor: s) findSolutionWithUnifier: Dictionary new in: termRelation.! !
Tree initialize!
